from typing import (
    Sequence,
)

from eth_utils import (
    to_tuple,
)

from ssz.sedes import (
    BaseSedes,
    Bitlist,
    Bitvector,
    Boolean,
    ByteVector,
    Container,
    List,
    Serializable,
    UInt,
    Vector,
)
from ssz.sedes.serializable import (
    MetaSerializable,
)

from .codec import (
    DefaultCodec,
)


def to_formatted_dict(value, sedes=None, codec=DefaultCodec):
    return dump(value, sedes, codec)


def dump(value, sedes=None, codec=DefaultCodec):
    if sedes is None:
        if isinstance(value, Serializable):
            return dump_serializable(value, codec)
        else:
            raise ValueError("must provide sedes for non-Serializable")

    if isinstance(sedes, Boolean):
        return dump_boolean(value, sedes, codec)
    elif isinstance(sedes, UInt):
        return dump_integer(value, sedes, codec)
    elif isinstance(sedes, ByteVector):
        return dump_bytes(value, sedes, codec)
    elif isinstance(sedes, List):
        return dump_list(value, sedes, codec)
    elif isinstance(sedes, Vector):
        return dump_vector(value, sedes, codec)
    elif isinstance(sedes, (Bitlist, Bitvector)):
        return dump_bits(value, sedes, codec)
    elif isinstance(sedes, Container):
        return dump_container(value, sedes, codec)
    elif isinstance(sedes, MetaSerializable):
        return dump_serializable(value, codec)
    elif isinstance(sedes, BaseSedes):
        raise Exception(f"Unreachable: All sedes types have been checked, {sedes} was not found")
    else:
        raise TypeError(f"Expected BaseSedes, got {type(sedes)}")


def dump_boolean(value, sedes, codec):
    if not isinstance(value, bool):
        raise ValueError(f"Expected value of type bool, got {type(value)}")
    return codec.encode_bool(value, sedes)


def dump_integer(value, sedes, codec):
    return codec.encode_integer(value, sedes)


def dump_bytes(value, sedes, codec):
    return codec.encode_bytes(value, sedes)


def dump_list(value, sedes, codec):
    if not isinstance(value, Sequence):
        raise ValueError(f"Expected Sequence, got {type(value)}")
    return tuple(dump(element, sedes.element_sedes, codec) for element in value)


def dump_vector(value, sedes, codec):
    if not isinstance(value, Sequence):
        raise ValueError(f"Expected Sequence, got {type(value)}")
    if not len(value) == sedes.length:
        raise ValueError(f"Expected {sedes.length} elements, got {len(value)}")
    return tuple(dump(element, sedes.element_sedes, codec) for element in value)


def dump_bits(value, sedes, codec):
    return codec.encode_bytes(sedes.serialize(value), sedes)


@to_tuple
def dump_container(value, sedes, codec):
    if not isinstance(value, Sequence):
        raise ValueError(f"Expected Sequence, got {type(value)}")
    elif not len(value) == len(sedes.field_sedes):
        raise ValueError(f"Expected {len(sedes.field_sedes)} elements, got {len(value)}")

    for element, element_sedes in zip(value, sedes.field_sedes):
        yield dump(
            element,
            element_sedes,
            codec,
        )


def dump_serializable(value, codec):
    dumped_values = dump(tuple(value), value._meta.container_sedes)
    return dict(zip(value._meta.field_names, dumped_values))



from .cache import (  # noqa: F401
    SSZCache,
)
from .codec import (  # noqa: F401
    decode,
    encode,
)
from .exceptions import (  # noqa: F401
    DeserializationError,
    SerializationError,
    SSZException,
)
from .sedes import (  # noqa: F401
    BaseSedes,
    BasicSedes,
    Bitlist,
    Bitvector,
    Boolean,
    Byte,
    ByteVector,
    CompositeSedes,
    Container,
    List,
    Serializable,
    SignedSerializable,
    UInt,
    Vector,
    boolean,
    byte_vector,
    bytes4,
    bytes32,
    bytes48,
    bytes96,
    empty_list,
    uint8,
    uint16,
    uint32,
    uint64,
    uint128,
    uint256,
)
from .tree_hash import (  # noqa: F401
    get_hash_tree_root,
)



import functools
from typing import (
    Any,
    Iterable,
)

from eth_typing import (
    Hash32,
)
from eth_utils import (
    to_tuple,
)

from ssz.sedes.base import (
    TSedes,
)
from ssz.typing import (
    CacheObj,
)


def get_key(sedes: TSedes, value: Any) -> str:
    key = _get_key(sedes, value).hex()
    sedes_name = type(sedes).__name__
    if len(key) > 0:
        return sedes_name + key
    else:
        # If the serialized result is empty, use sedes name as the key
        if hasattr(sedes, 'element_sedes'):
            return sedes_name + str(sedes.max_length)
        else:
            return sedes_name


@functools.lru_cache(maxsize=2**12)
def _get_key(sedes: TSedes, value: Any) -> bytes:
    return sedes.serialize(value)


@to_tuple
def get_merkle_leaves_without_cache(value: Any, element_sedes: TSedes) -> Iterable[Hash32]:
    for element in value:
        yield element_sedes.get_hash_tree_root(element)


@to_tuple
def get_merkle_leaves_with_cache(value: Any,
                                 element_sedes: TSedes,
                                 cache: CacheObj) -> Iterable[Hash32]:
    """
    NOTE: cache is mutable
    """
    for element in value:
        key = element_sedes.get_key(element)
        if key not in cache:
            root, cache = (
                element_sedes.get_hash_tree_root_and_leaves(
                    element,
                    cache,
                )
            )
            cache[key] = root
        yield cache[key]



from collections.abc import (
    Mapping,
)
from typing import (
    Sequence,
)

from eth_utils import (
    to_tuple,
)

from ssz.sedes import (
    BaseSedes,
    Bitlist,
    Bitvector,
    Boolean,
    ByteVector,
    Container,
    List,
    UInt,
    Vector,
)
from ssz.sedes.serializable import (
    MetaSerializable,
)

from .codec import (
    DefaultCodec,
)


def from_formatted_dict(value, sedes, codec=DefaultCodec):
    return parse(value, sedes, codec)


def parse(value, sedes, codec=DefaultCodec):
    if isinstance(sedes, Boolean):
        return parse_boolean(value, sedes, codec)
    elif isinstance(sedes, UInt):
        return parse_integer(value, sedes, codec)
    elif isinstance(sedes, ByteVector):
        return parse_bytes(value, sedes, codec)
    elif isinstance(sedes, List):
        return parse_list(value, sedes, codec)
    elif isinstance(sedes, Vector):
        return parse_vector(value, sedes, codec)
    elif isinstance(sedes, Container):
        return parse_container(value, sedes, codec)
    elif isinstance(sedes, (Bitlist, Bitvector)):
        return parse_bits(value, sedes, codec)
    elif isinstance(sedes, MetaSerializable):
        return parse_serializable(value, sedes, codec)
    elif isinstance(sedes, BaseSedes):
        raise Exception(f"Unreachable: All sedes types have been checked, {sedes} was not found")
    else:
        raise TypeError(f"Expected BaseSedes, got {type(sedes)}")


def parse_boolean(value, sedes, codec):
    if not isinstance(value, bool):
        raise ValueError(f"Expected value of type bool, got {type(value)}")
    return codec.decode_bool(value, sedes)


def parse_integer(value, sedes, codec):
    return codec.decode_integer(value, sedes)


def parse_bytes(value, sedes, codec):
    return codec.decode_bytes(value, sedes)


def parse_list(value, sedes, codec):
    if not isinstance(value, Sequence):
        raise ValueError(f"Expected Sequence, got {type(value)}")
    return tuple(parse(element, sedes.element_sedes, codec) for element in value)


def parse_vector(value, sedes, codec):
    if not isinstance(value, Sequence):
        raise ValueError(f"Expected Sequence, got {type(value)}")
    if not len(value) == sedes.length:
        raise ValueError(f"Expected {sedes.length} elements, got {len(value)}")
    return tuple(parse(element, sedes.element_sedes, codec) for element in value)


def parse_bits(value, sedes, codec):
    if not isinstance(value, Sequence):
        raise ValueError(f"Expected Sequence, got {type(value)}")

    data = parse_bytes(value, sedes, codec)
    deserialized = sedes.deserialize(data)
    return tuple(parse_boolean(element, sedes, codec) for element in deserialized)


@to_tuple
def parse_container(value, sedes, codec):
    if not isinstance(value, Sequence):
        raise ValueError(f"Expected Sequence, got {type(value)}")
    elif not len(value) == len(sedes.field_sedes):
        raise ValueError(f"Expected {len(sedes.field_sedes)} elements, got {len(value)}")

    for element, element_sedes in zip(value, sedes.field_sedes):
        yield parse(
            element,
            element_sedes,
            codec,
        )


def parse_serializable(value, serializable_cls, codec):
    if not isinstance(value, Mapping):
        raise ValueError(f"Expected Mapping, got {type(value)}")
    parse_args = tuple(value[field_name] for field_name in serializable_cls._meta.field_names)
    input_args = parse(parse_args, serializable_cls._meta.container_sedes)
    input_kwargs = dict(zip(serializable_cls._meta.field_names, input_args))
    return serializable_cls(**input_kwargs)



from typing import (
    Any,
)

from eth_typing import (
    Hash32,
)

from ssz.sedes import (
    infer_sedes,
)
from ssz.sedes.base import (
    BaseSedes,
)


def get_hash_tree_root(value: Any, sedes: BaseSedes = None) -> Hash32:
    if sedes is None:
        sedes = infer_sedes(value)

    return sedes.get_hash_tree_root(value)



from eth_utils import (
    decode_hex,
    encode_hex,
)


class DefaultCodec:

    @staticmethod
    def encode_integer(value: int, sedes):
        return value

    @staticmethod
    def encode_bool(value: bool, sedes):
        return value

    @staticmethod
    def encode_bytes(value: bytes, sedes):
        return encode_hex(value)

    @staticmethod
    def decode_bool(value, sedes) -> bool:
        return value

    @staticmethod
    def decode_integer(value, sedes) -> int:
        if not isinstance(value, int):
            raise ValueError(f"Expected value of type int, got {type(value)}")
        return value

    @staticmethod
    def decode_bytes(value, sedes) -> bytes:
        return decode_hex(value)



from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    Any,
    Generic,
    Tuple,
)

from eth_typing import (
    Hash32,
)

from ssz.typing import (
    CacheObj,
    TDeserialized,
    TSerializable,
)


class BaseSedes(ABC, Generic[TSerializable, TDeserialized]):
    #
    # Size
    #
    @property
    @abstractmethod
    def is_fixed_sized(self) -> bool:
        ...

    @abstractmethod
    def get_fixed_size(self) -> int:
        ...

    #
    # Serialization
    #
    @abstractmethod
    def serialize(self, value: TSerializable) -> bytes:
        ...

    #
    # Deserialization
    #
    @abstractmethod
    def deserialize(self, data: bytes) -> TDeserialized:
        ...

    #
    # Tree hashing
    #
    @abstractmethod
    def get_hash_tree_root(self, value: TSerializable) -> Hash32:
        ...

    @abstractmethod
    def get_hash_tree_root_and_leaves(self,
                                      value: TSerializable,
                                      cache: CacheObj) -> Tuple[Hash32, CacheObj]:
        ...

    @abstractmethod
    def chunk_count(self) -> int:
        ...

    @abstractmethod
    def get_key(self, value: Any) -> bytes:
        ...


TSedes = BaseSedes[Any, Any]


class BaseCompositeSedes(BaseSedes[TSerializable, TDeserialized]):
    @abstractmethod
    def get_key(self, value: Any) -> bytes:
        ...


class BaseByteSedes(BaseSedes[TSerializable, TDeserialized]):
    @abstractmethod
    def get_key(self, value: Any) -> bytes:
        ...



from typing import (
    Sequence,
    Tuple,
    Union,
)

from eth_typing import (
    Hash32,
)
from eth_utils import (
    to_tuple,
)

from ssz.exceptions import (
    DeserializationError,
    SerializationError,
)
from ssz.sedes.basic import (
    BasicBytesSedes,
)
from ssz.typing import (
    CacheObj,
)
from ssz.utils import (
    get_serialized_bytearray,
    merkleize,
    merkleize_with_cache,
    mix_in_length,
    pack_bits,
)

BytesOrByteArray = Union[bytes, bytearray]


class Bitlist(BasicBytesSedes[BytesOrByteArray, bytes]):
    def __init__(self, max_bit_count: int) -> None:
        if max_bit_count < 0:
            raise TypeError("Max bit count cannot be negative")
        self.max_bit_count = max_bit_count

    #
    # Size
    #
    is_fixed_sized = False

    def get_fixed_size(self):
        raise ValueError("byte list has no static size")

    #
    # Serialization
    #
    def serialize(self, value: Sequence[bool]) -> bytes:
        len_value = len(value)
        if len_value > self.max_bit_count:
            raise SerializationError(
                f"Cannot serialize length {len_value} bit array as Bitlist[{self.max_bit_count}]"
            )

        serialized_bytearray = get_serialized_bytearray(value, len_value, extra_byte=True)
        serialized_bytearray[len_value // 8] |= 1 << (len_value % 8)
        return bytes(serialized_bytearray)

    #
    # Deserialization
    #
    @to_tuple
    def deserialize(self, data: bytes) -> Tuple[bool, ...]:
        as_integer = int.from_bytes(data, 'little')
        len_value = get_bitlist_len(as_integer)

        if len_value > self.max_bit_count:
            raise DeserializationError(
                f"Cannot deserialize length {len_value} bytes data as Bitlist[{self.max_bit_count}]"
            )

        for bit_index in range(len_value):
            yield bool((data[bit_index // 8] >> bit_index % 8) % 2)

    #
    # Tree hashing
    #
    def get_hash_tree_root(self, value: Sequence[bool]) -> bytes:
        return mix_in_length(merkleize(pack_bits(value), limit=self.chunk_count()), len(value))

    def get_hash_tree_root_and_leaves(self,
                                      value: Sequence[bool],
                                      cache: CacheObj) -> Tuple[Hash32, CacheObj]:
        root, cache = merkleize_with_cache(
            pack_bits(value),
            cache=cache,
            limit=self.chunk_count(),
        )
        return mix_in_length(root, len(value)), cache

    def chunk_count(self) -> int:
        return (self.max_bit_count + 255) // 256


def get_bitlist_len(x: int) -> int:
    return x.bit_length() - 1



from collections.abc import (
    MutableMapping,
)
from typing import (
    NewType,
    TypeVar,
    Union,
)

TSerializable = TypeVar("TSerializable")
TDeserialized = TypeVar("TDeserialized")

TSerializableElement = TypeVar("TSerializable")
TDeserializedElement = TypeVar("TDeserialized")

CacheObj = NewType('CacheObj', Union[MutableMapping, dict])



from collections.abc import (
    Iterable,
)

from .base import (  # noqa: F401
    BaseSedes,
)
from .basic import (  # noqa: F401
    BasicSedes,
    CompositeSedes,
)
from .bitlist import (  # noqa: F401
    Bitlist,
)
from .bitvector import (  # noqa: F401
    Bitvector,
)
from .boolean import (  # noqa: F401
    Boolean,
    boolean,
)
from .byte import (  # noqa: F401
    Byte,
    byte,
)
from .byte_vector import (  # noqa: F401
    ByteVector,
    bytes1,
    bytes4,
    bytes32,
    bytes48,
    bytes96,
)
from .container import (  # noqa: F401
    Container,
)
from .list import (  # noqa: F401
    List,
    empty_list,
)
from .serializable import (  # noqa: F401
    Serializable,
)
from .signed_serializable import (  # noqa: F401
    SignedSerializable,
)
from .uint import (  # noqa: F401
    UInt,
    uint8,
    uint16,
    uint32,
    uint64,
    uint128,
    uint256,
)
from .vector import (  # noqa: F401
    Vector,
)

sedes_by_name = {
    "bool": boolean,

    "byte": byte,
    "bytes4": bytes4,
    "bytes32": bytes32,
    "bytes48": bytes48,
    "bytes96": bytes96,
    "empty_list": empty_list,

    "uint8": uint8,
    "uint16": uint16,
    "uint32": uint32,
    "uint64": uint64,
    "uint128": uint128,
    "uint256": uint256,
}


def infer_sedes(value):
    """
    Try to find a sedes objects suitable for a given Python object.
    """
    if isinstance(value.__class__, BaseSedes):
        return value.__class__  # Mainly used for `Serializable` classes
    elif isinstance(value, bool):
        return boolean
    elif isinstance(value, int):
        raise TypeError("uint sedes object or uint string needs to be specified for ints")
    elif isinstance(value, Iterable):
        raise TypeError("Cannot infer list sedes for iterables that are not sequences")
    else:
        raise TypeError(f"Did not find sedes handling type {type(value).__name__}")



from collections.abc import (
    MutableMapping,
)
from typing import (
    TYPE_CHECKING,
    Iterator,
)

from lru import LRU

if TYPE_CHECKING:
    MM = MutableMapping[bytes, bytes]
else:
    MM = MutableMapping

DEFAULT_CACHE_SIZE = 2**10


class SSZCache(MM):
    def __init__(self, cache_size: int=DEFAULT_CACHE_SIZE) -> None:
        self._cache_size = cache_size
        self.clear()

    def clear(self) -> None:
        self._cached_values = LRU(self._cache_size)

    def _exists(self, key: bytes) -> bool:
        return key in self._cached_values

    def __contains__(self, key: bytes) -> bool:
        return self._exists(key)

    def __getitem__(self, key: bytes) -> bytes:
        return self._cached_values[key]

    def __setitem__(self, key: bytes, value: bytes) -> None:
        self._cached_values[key] = value

    def __delitem__(self, key: bytes) -> None:
        if key in self._cached_values:
            del self._cached_values[key]
        else:
            raise KeyError(f"key: {key} not found")

    def __iter__(self) -> Iterator[bytes]:
        raise NotImplementedError("By default, DB classes cannot be iterated.")

    def __len__(self) -> int:
        raise NotImplementedError("By default, classes cannot return the total number of keys.")

    @property
    def cache_size(self) -> int:
        return self._cache_size



from typing import (
    Tuple,
    Union,
)

from eth_typing import (
    Hash32,
)

from ssz.exceptions import (
    DeserializationError,
    SerializationError,
)
from ssz.sedes.basic import (
    BasicBytesSedes,
)
from ssz.typing import (
    CacheObj,
)
from ssz.utils import (
    merkleize,
    merkleize_with_cache,
    pack_bytes,
)

BytesOrByteArray = Union[bytes, bytearray]


class ByteVector(BasicBytesSedes[BytesOrByteArray, bytes]):
    def __init__(self, size: int) -> None:
        if size < 0:
            raise TypeError("Size cannot be negative")
        self.size = size

    #
    # Size
    #
    is_fixed_sized = True

    def get_fixed_size(self):
        return self.size

    #
    # Serialization
    #
    def serialize(self, value: BytesOrByteArray) -> bytes:
        if len(value) != self.size:
            raise SerializationError(
                f"Cannot serialize length {len(value)} byte-string as bytes{self.size}"
            )

        return value

    #
    # Deserialization
    #
    def deserialize(self, data: bytes) -> bytes:
        if len(data) != self.size:
            raise DeserializationError(
                f"Cannot deserialize length {len(data)} data as bytes{self.size}"
            )
        return data

    #
    # Tree hashing
    #
    def get_hash_tree_root(self, value: bytes) -> bytes:
        serialized_value = self.serialize(value)
        return merkleize(pack_bytes(serialized_value))

    def get_hash_tree_root_and_leaves(self,
                                      value: bytes,
                                      cache: CacheObj) -> Tuple[Hash32, CacheObj]:
        serialized_value = self.serialize(value)
        return merkleize_with_cache(
            pack_bytes(serialized_value),
            cache=cache,
        )

    def chunk_count(self) -> int:
        return self.length * self.size


bytes1 = ByteVector(1)
bytes4 = ByteVector(4)
bytes8 = ByteVector(8)
bytes32 = ByteVector(32)
bytes48 = ByteVector(48)
bytes96 = ByteVector(96)
